import sys
from typing import List, Tuple, TYPE_CHECKING

if TYPE_CHECKING:
    from importlib.metadata import EntryPoint, Distribution

if sys.version_info >= (3,10,0):
    from importlib import metadata

    def _get_entry_points(group_name: str) -> List[Tuple['EntryPoint', 'Distribution']]:
        eps = []
        for ep in metadata.entry_points().select(group=group_name):
            eps.append((ep, ep.dist))
        return eps

    def _get_name_from_dist(dist: 'Distribution') -> str:
        return dist.name

elif sys.version_info >= (3,8,0):
    from importlib import metadata

    def _get_entry_points(group_name: str) -> List[Tuple['EntryPoint', 'Distribution']]:
        eps = []
        for dist in metadata.distributions():
            for ep in dist.entry_points:
                if ep.group == group_name:
                    eps.append((ep, dist))
        return eps

    def _get_name_from_dist(dist: 'Distribution') -> str:
        return dist.metadata["Name"]

else:
    import pkg_resources # type: ignore

    def _get_entry_points(group_name: str) -> List[Tuple['EntryPoint', 'Distribution']]:
        eps = []
        for ep in pkg_resources.iter_entry_points(group_name):
            eps.append((ep, ep.dist))
        return eps

    def _get_name_from_dist(dist: 'Distribution') -> str:
        return dist.project_name # type: ignore


def get_entry_points(group_name: str) -> List[Tuple['EntryPoint', 'Distribution']]:
    return _get_entry_points(group_name)

def get_name_from_dist(dist: 'Distribution') -> str:
    return _get_name_from_dist(dist)



from typing import TYPE_CHECKING, List, Dict, Any
import re
import os

from systemrdl.messages import FileSourceRef

if TYPE_CHECKING:
    import argparse
    from typing import Sequence
    from systemrdl import RDLCompiler, AddrmapNode
    from .importer import Importer


def add_rdl_compile_arguments(parser: 'argparse._ActionsContainer') -> None:
    parser.add_argument(
        "input_files",
        metavar="FILE",
        nargs="+",
        help="One or more input files"
    )
    parser.add_argument(
        "-I",
        dest="incdirs",
        metavar="INCDIR",
        action="append",
        help='Search directory for files included with `include "filename"',
    )


def add_importer_arguments(parser: 'argparse._ActionsContainer', importers: 'Sequence[Importer]') -> None:
    for importer in importers:
        importer_arg_group = parser.add_argument_group(f"{importer.name} importer args")
        importer.add_importer_arguments(importer_arg_group)


def add_elaborate_arguments(parser: 'argparse._ActionsContainer') -> None:
    parser.add_argument(
        "-t", "--top",
        dest="top_def_name",
        metavar="TOP",
        default=None,
        help="Explicitly choose which addrmap  in the root namespace will be the "
                "top-level component. If unset, The last addrmap defined will be chosen"
    )
    parser.add_argument(
        "--rename",
        dest="inst_name",
        default=None,
        help="Overrides the top-component's instantiated name. By default, the "
                "instantiated name is the same as the component's type name"
    )
    parser.add_argument(
        "-P",
        dest="parameters",
        metavar="PARAMETER=VALUE",
        action="append",
        default=[],
        help='Specify value for a top-level SystemRDL parameter',
    )


def parse_parameters(rdlc: 'RDLCompiler', parameter_options: List[str]) -> Dict[str, Any]:
    parameters = {}
    for raw_param in parameter_options:
        m = re.fullmatch(r"(\w+)=(.+)", raw_param)
        if not m:
            rdlc.msg.fatal(f"Invalid parameter argument: {raw_param}")

        p_name = m.group(1)
        try:
            p_value = rdlc.eval(m.group(2))
        except ValueError:
            rdlc.msg.fatal(f"Unable to parse value '{m.group(2)}' for parameter '{p_name}'")
        parameters[p_name] = p_value

    return parameters


def process_input(rdlc: 'RDLCompiler', importers: 'Sequence[Importer]', input_files: List[str], options: 'argparse.Namespace') -> None:
    for file in input_files:
        if not os.path.exists(file):
            rdlc.msg.fatal(f"Input file does not exist: {file}")

        ext = os.path.splitext(file)[1].strip(".")
        if ext == "rdl":
            # Is SystemRDL file
            rdlc.compile_file(
                file,
                incl_search_paths=options.incdirs,
            )
        else:
            # Is foreign input file.

            # Search which importer to use by extension first
            importer_candidates = [] # type: List[Importer]
            for importer in importers:
                if ext in importer.file_extensions:
                    importer_candidates.append(importer)

            # Do 2nd pass if needed
            importer = None
            if len(importer_candidates) == 1:
                importer = importer_candidates[0]
            elif len(importer_candidates) > 1:
                # ambiguous which importer to use
                # Do 2nd pass compatibility check
                for importer_candidate in importer_candidates:
                    if importer_candidate.is_compatible(file):
                        importer = importer_candidate
                        break

            if not importer:
                rdlc.msg.fatal(
                    "Unknown file type. Could not find any importers capable of reading this file.",
                    FileSourceRef(file)
                )

            importer.do_import(rdlc, options, file)


def elaborate(rdlc: 'RDLCompiler', parameters: Dict[str, Any], options: 'argparse.Namespace') -> 'AddrmapNode':
    try:
        root = rdlc.elaborate(
            top_def_name=options.top_def_name,
            inst_name=options.inst_name,
            parameters=parameters
        )
    except (ValueError, TypeError) as e:
        # Parameter issues raise ValueError or TypeError
        # TODO: Fix exception types once they become specialized in the compiler
        rdlc.msg.fatal(e.args[0])

    return root.top



from typing import TYPE_CHECKING

from .config import schema
from .config.loader import AppConfig

if TYPE_CHECKING:
    from typing import List, Dict, Any
    import argparse
    from systemrdl import RDLCompiler

class Importer:
    # Importer name
    name = None # type: str

    #: A list of one or more file extensions the importer expects to support.
    #: This is used as a rough first-pass method to identify which
    #: importer is appropriate for a given file type.
    file_extensions = [] # type: List[str]

    #: Schema for additional organization-specific configuration options
    #: specified by a 'peakrdl.toml' file loaded at startup
    #:
    #: For more details, see :ref:`cfg_schema`
    cfg_schema = {} # type: Dict[str, Any]

    def __init__(self) -> None:
        #: Resolved configuration data that was extracted from the PeakRDL TOML,
        #: and validated.
        self.cfg = {} # type: Dict[str, Any]

    def _load_cfg(self, cfg: AppConfig) -> None:
        self.cfg = cfg.get_namepsace(self.name, schema.normalize(self.cfg_schema))

    def is_compatible(self, path: str) -> bool:
        """
        This function is used to further determine if this importer is capable of
        processing the given input file.

        If the file extension was not enough to determine which importer to use,
        this function is called and shall determine if the file is compatible
        with this importer to a high degree of confidence.

        .. note::

            This should not attempt to exaustively validate the file's
            correctness.

            Instead, it is recommended to open the file, and perform a low-cost
            scan of the contents to quickly determine if the file's format
            appears to be compatible with this importer.
            This can be as simple as a quick keyword search.

        Parameters
        ----------
        path: str
            Path to the input file
        """
        raise NotImplementedError


    def add_importer_arguments(self, arg_group: 'argparse._ActionsContainer') -> None:
        """
        Override this function to define additional command line arguments by
        using the ``arg_group.add_argument()`` method.
        See Python's `argparse module <https://docs.python.org/3/library/argparse.html#the-add-argument-method>`_
        for more details on how to use this.

        .. note::

            Not all exporter configuration options are appropriate as command-line arguments.
            For options that will likely remain static for a given user/organization,
            consider using the PeakRDL TOML configuration mechanism via the
            ``cft_schema`` and ``cfg`` class members.

        Parameters
        ----------
        arg_group: ``argparse.ArgumentParser``
            Add more command line arguments via this object.
        """


    def do_import(self, rdlc: 'RDLCompiler', options: 'argparse.Namespace', path: str) -> None:
        """
        Defines the implementation of your importer.

        Parameters
        ----------
        rdlc: ``systemrdl.RDLCompiler``
            Reference to the SystemRDL ``RDLCompiler`` object.
        options: ``argparse.Namespace``
            Argparse Namespace object containing all the command line argument values
        path: str
            Path to the input file
        """
        raise NotImplementedError



from typing import TYPE_CHECKING
import math

from systemrdl import RDLListener, RDLWalker
from systemrdl.node import AddrmapNode, RegNode

from ..subcommand import ExporterSubcommand

if TYPE_CHECKING:
    import argparse


class DumpListener(RDLListener):
    def __init__(self, hex_digits:int, unroll: bool) -> None:
        self.hex_digits = hex_digits
        self.unroll = unroll

    def enter_Reg(self, node: RegNode) -> None:
        if self.unroll:
            addr = node.absolute_address
            size = node.size
        else:
            addr = node.raw_absolute_address
            size = node.total_size

        print(
            f"0x{addr:0{self.hex_digits}x}-0x{addr+size-1:0{self.hex_digits}x}:",
            node.get_path(empty_array_suffix="[{dim:d}]")
        )


class Dump(ExporterSubcommand):
    name = "dump"
    short_desc = "print register model contents to stdout"
    generates_output_file = False

    def add_exporter_arguments(self, arg_group: 'argparse._ActionsContainer') -> None:

        arg_group.add_argument(
            "-u", "--unroll",
            default=False,
            action="store_true",
            help="Unroll arrays"
        )


    def do_export(self, top_node: AddrmapNode, options: 'argparse.Namespace') -> None:
        hex_digits = math.ceil(top_node.total_size.bit_length() / 4)
        walker = RDLWalker(unroll=options.unroll)
        listener = DumpListener(hex_digits, options.unroll)
        walker.walk(top_node, listener)



import argparse
import sys
import os
import shlex
from typing import TYPE_CHECKING, List, Dict, Optional, Set

from systemrdl import RDLCompileError

from .__about__ import __version__
from .config.loader import load_cfg, AppConfig
from .plugins.exporter import get_exporter_plugins
from .plugins.importer import get_importer_plugins
from .cmd.dump import Dump
from .cmd.list_globals import ListGlobals
from .cmd.preprocess import Preprocess


if TYPE_CHECKING:
    from .subcommand import Subcommand


DESCRIPTION = """
PeakRDL is a control & status register model automation toolchain.

For help about a specific subcommand, try:
    peakrdl <command> --help

For more documentation, visit https://peakrdl.readthedocs.io

"""

class SubcommandHelpFormatter(argparse.RawDescriptionHelpFormatter):
    def _format_action(self, action): # type: ignore
        parts = super(argparse.RawDescriptionHelpFormatter, self)._format_action(action)
        if action.nargs == argparse.PARSER:
            parts = "\n".join(parts.split("\n")[1:])
        return parts


class ReportPluginsImpl(argparse.Action):
    CFG = None # type: AppConfig

    def __call__ (self, parser, namespace, values, option_string = None): # type: ignore
        exporters = get_exporter_plugins(self.CFG)
        importers = get_importer_plugins(self.CFG)

        print("importers:")
        for importer in importers:
            print(f"\t{importer.plugin_info}")
        print("exporters:")
        for exporter in exporters:
            print(f"\t{exporter.plugin_info}")
        sys.exit(0)


def get_file_args(path: str) -> List[str]:
    if not os.path.exists(path):
        print(f"error: file not found: {path}", file=sys.stderr)
        sys.exit(1)

    with open(path, "r", encoding='utf-8') as f:
        return shlex.split(f.read(), comments=True)


def expand_file_args(argv: List[str], _pathlist: Optional[Set[str]] = None) -> List[str]:
    if _pathlist is None:
        _pathlist = set()

    new_argv = []
    argv_iter = iter(argv)
    for arg in argv_iter:
        if arg == "-f":
            try:
                path = next(argv_iter)
            except StopIteration:
                print("error: argument -f: expected FILE", file=sys.stderr)
                sys.exit(1)

            if path in _pathlist:
                print(f"error: circular reference in -f files: '{path}' was already opened", file=sys.stderr)
                sys.exit(1)
            _pathlist.add(path)
            file_args = get_file_args(path)
            file_args = expand_file_args(file_args, _pathlist)
            _pathlist.remove(path)
            new_argv.extend(file_args)
        else:
            new_argv.append(arg)
    return new_argv


def main() -> None:
    # manually expand any -f argfiles first
    argv = expand_file_args(sys.argv[1:])

    cfg = load_cfg(argv)

    # Collect all importers and initialize them with the config
    importers = get_importer_plugins(cfg)
    for importer in importers:
        importer._load_cfg(cfg)

    # Collect all subcommands
    subcommands = [
        Dump(),
        ListGlobals(),
        Preprocess(),
    ] # type: List[Subcommand]
    subcommands += get_exporter_plugins(cfg)
    for subcommand in subcommands:
        subcommand._load_cfg(cfg)

    # Check for duplicate subcommands
    sc_dict = {} # type: Dict[str, Subcommand]
    for sc in subcommands:
        if sc.name in sc_dict:
            raise RuntimeError(f"More than one exporter plugin was registered with the same name '{sc.name}': \n\t{sc_dict[sc.name]}\n\t{sc}")
        sc_dict[sc.name] = sc

    # Initialize top-level arg parser
    class ReportPlugins(ReportPluginsImpl):
        CFG = cfg
    parser = argparse.ArgumentParser(
        description=DESCRIPTION,
        formatter_class=SubcommandHelpFormatter,
    )
    parser.add_argument("--version", action="version", version=__version__)
    parser.add_argument(
        "--plugins", action=ReportPlugins, nargs=0,
        help="Report the PeakRDL plugins, their versions, then exit"
    )

    # Add dummy -f and cfg flags. Not actually used as these are already
    # expanded earlier manually
    parser.add_argument(
        '-f',
        metavar="FILE",
        dest="argfile",
        help="Specify a file containing more command line arguments"
    )
    parser.add_argument(
        '--peakrdl-cfg',
        metavar="CFG",
        dest="peakrdl_cfg",
        help="Specify a PeakRDL configuration TOML file"
    )

    # Initialize subcommand arg parsers
    subgroup = parser.add_subparsers(
        title="subcommands",
        metavar="<subcommand>",
    )
    for subcommand in subcommands:
        subcommand._init_subparser(subgroup, importers)

    # Process command-line args
    options = parser.parse_args(argv)
    if not hasattr(options, 'subcommand'):
        parser.print_usage()
        print(f"{parser.prog}: error the following arguments are required: <subcommand>")
        sys.exit(1)

    # Run subcommand!
    try:
        options.subcommand.main(importers, options)
    except RDLCompileError:
        sys.exit(1)



from typing import TYPE_CHECKING, List

from systemrdl import RDLCompiler

from ..subcommand import Subcommand

if TYPE_CHECKING:
    import argparse
    from ..plugins.importer import ImporterPlugin

class Preprocess(Subcommand):
    name = "preprocess"
    short_desc = "Preprocess SystemRDL and write the result to a file"

    def add_arguments(self, parser: 'argparse._ActionsContainer', importers: 'List[ImporterPlugin]') -> None:
        grp = parser.add_argument_group("preprocessor args")
        grp.add_argument(
            "file",
            help="SystemRDL file to preprocess"
        )
        grp.add_argument(
            "-I",
            dest="incdirs",
            metavar="INCDIR",
            action="append",
            help='Search directory for files included with `include "filename"',
        )
        grp.add_argument(
            "-o",
            dest="output",
            required=True,
            help="Output path",
        )

    def main(self, importers: 'List[ImporterPlugin]', options: 'argparse.Namespace') -> None:
        rdlc = RDLCompiler()
        f_info = rdlc.preprocess_file(options.file, options.incdirs)
        with open(options.output, 'w', encoding='utf-8') as f:
            f.write(f_info.preprocessed_text)



from typing import TYPE_CHECKING, Optional, List, Type, Dict, Any

from systemrdl import RDLCompiler

from .config import schema
from .config.loader import AppConfig
from . import process_input

if TYPE_CHECKING:
    import argparse

    from systemrdl.node import AddrmapNode
    from systemrdl.udp import UDPDefinition
    from .plugins.importer import ImporterPlugin

class Subcommand:
    """
    Base command line interface subcommand class
    """

    # Subcommand name
    name = None # type: str

    #: A brief one-line description of the exporter command.
    short_desc = None # type: str

    #: Longer-form description.
    #: If left as None, inherits short_desc
    long_desc = None # type: Optional[str]

    #: Schema for additional organization-specific configuration options
    #: specified by a 'peakrdl.toml' file loaded at startup.
    #:
    #: For more details, see :ref:`cfg_schema`
    cfg_schema = {} # type: Dict[str, Any]

    def __init__(self) -> None:
        #: Resolved configuration data that was extracted from the PeakRDL TOML,
        #: and validated.
        self.cfg = {} # type: Dict[str, Any]

    def _load_cfg(self, cfg: AppConfig) -> None:
        self.cfg = cfg.get_namepsace(self.name, schema.normalize(self.cfg_schema))

    def _init_subparser(self, subgroup: 'argparse._SubParsersAction', importers: 'List[ImporterPlugin]') -> None:
        assert isinstance(self.name, str)
        assert isinstance(self.short_desc, str)
        subparser = subgroup.add_parser(
            self.name,
            help=self.short_desc,
            description=(self.long_desc or self.short_desc)
        )
        self.add_arguments(subparser, importers)
        subparser.set_defaults(subcommand=self)

        # Add dummy -f and cfg flags. Not actually used as these are already
        # expanded earlier manually
        subparser.add_argument(
            '-f',
            metavar="FILE",
            dest="argfile",
            help="Specify a file containing more command line arguments"
        )
        subparser.add_argument(
            '--peakrdl-cfg',
            metavar="CFG",
            dest="peakrdl_cfg",
            help="Specify a PeakRDL configuration TOML file"
        )


    def add_arguments(self, parser: 'argparse._ActionsContainer', importers: 'List[ImporterPlugin]') -> None:
        pass


    def main(self, importers: 'List[ImporterPlugin]', options: 'argparse.Namespace') -> None:
        raise NotImplementedError



class ExporterSubcommand(Subcommand):
    """
    Basic PeakRDL exporter subcommand.
    Most subcommands will fall under this category as they do the following:
    - Compile one or more RDL files
    - Optionally import other non-RDL sources
    - Elaborate the register model
    - Export <something>
    """

    #: Determines whether this subcommand should require the user to provide an
    #: output path. If ``True``, adds a required ``-o`` command-line argument.
    #: The result of this is available later in the ``do_export()`` function
    #: via ``options.output``.
    #:
    #: Set this to ``False`` if your exporter does not write any output files.
    generates_output_file = True

    #: List of ``systemrdl.udp.UDPDefinition`` classes that this subcommand
    #: provides. Internally, each of these definitions are registered with the
    #: compiler as soft UDPs via ``RDLCompiler.register_udp()``
    udp_definitions = [] # type: List[Type[UDPDefinition]]

    def add_arguments(self, parser: 'argparse._ActionsContainer', importers: 'List[ImporterPlugin]') -> None:
        compiler_arg_group = parser.add_argument_group("compilation args")
        process_input.add_rdl_compile_arguments(compiler_arg_group)
        process_input.add_elaborate_arguments(compiler_arg_group)

        process_input.add_importer_arguments(parser, importers)

        exporter_arg_group = parser.add_argument_group("exporter args")
        if self.generates_output_file:
            exporter_arg_group.add_argument(
                "-o",
                dest="output",
                required=True,
                help="Output path",
            )
        self.add_exporter_arguments(exporter_arg_group)

    def add_exporter_arguments(self, arg_group: 'argparse._ActionsContainer') -> None:
        """
        Override this function to define additional command line arguments by
        using the ``arg_group.add_argument()`` method.
        See Python's `argparse module <https://docs.python.org/3/library/argparse.html#the-add-argument-method>`_
        for more details on how to use this.

        .. note::

            Not all exporter configuration options are appropriate as command-line arguments.
            For options that will likely remain static for a given user/organization,
            consider using the PeakRDL TOML configuration mechanism via the
            ``cft_schema`` and ``cfg`` class members.

        Parameters
        ----------
        arg_group: ``argparse.ArgumentParser``
            Add more command line arguments via this object.
        """

    def main(self, importers: 'List[ImporterPlugin]', options: 'argparse.Namespace') -> None:
        rdlc = RDLCompiler()

        for udp in self.udp_definitions:
            rdlc.register_udp(udp)

        parameters = process_input.parse_parameters(rdlc, options.parameters)

        process_input.process_input(rdlc, importers, options.input_files, options)

        top = process_input.elaborate(rdlc, parameters, options)

        # Run exporter
        self.do_export(top, options)


    def do_export(self, top_node: 'AddrmapNode', options: 'argparse.Namespace') -> None:
        """
        Override this function to define the implementation of your exporter.

        Parameters
        ----------
        top_node: ``systemrdl.node.AddrmapNode``
            Node representing the top of the design to be exported
        options: ``argparse.Namespace``
            Argparse namespace object containing all the command line argument values.
        """
        raise NotImplementedError



from typing import TYPE_CHECKING, List

from systemrdl import RDLCompiler
from systemrdl.component import Addrmap

from ..subcommand import Subcommand
from .. import process_input

if TYPE_CHECKING:
    import argparse
    from ..plugins.importer import ImporterPlugin

class ListGlobals(Subcommand):
    name = "globals"
    short_desc = "list all globally accessible types that can be elaborated as top"

    def add_arguments(self, parser: 'argparse._ActionsContainer', importers: 'List[ImporterPlugin]') -> None:
        compiler_arg_group = parser.add_argument_group("compilation args")
        process_input.add_rdl_compile_arguments(compiler_arg_group)

        process_input.add_importer_arguments(parser, importers)

    def main(self, importers: 'List[ImporterPlugin]', options: 'argparse.Namespace') -> None:
        rdlc = RDLCompiler()
        process_input.process_input(rdlc, importers, options.input_files, options)

        for name, comp_def in rdlc.root.comp_defs.items():
            if isinstance(comp_def, Addrmap):
                print(name)



from typing import Any, List, Dict, Union
import datetime
import os
import re
import importlib

class SchemaException(Exception):
    """
    Raised if encounters a data extraction error.
    """

class Schema:
    def extract(self, data: Any, path: str, err_ctx: str) -> Any:
        raise NotImplementedError

RawSchema = Union[Schema, list, dict]

#-------------------------------------------------------------------------------
# Base TOML datatypes
#-------------------------------------------------------------------------------
class _SimpleType(Schema):
    TYPE = object
    def extract(self, data: Any, path: str, err_ctx: str) -> Any:
        if not isinstance(data, self.TYPE):
            raise SchemaException(f"{err_ctx}: Expected {self.TYPE.__name__}. Got {type(data).__name__}")
        return data

class String(_SimpleType):
    TYPE = str

class Integer(_SimpleType):
    TYPE = int

class Float(_SimpleType):
    TYPE = float

class Boolean(_SimpleType):
    TYPE = bool

class DateTime(_SimpleType):
    TYPE = datetime.datetime # type: ignore

class Date(_SimpleType):
    TYPE = datetime.date # type: ignore

class Time(_SimpleType):
    TYPE = datetime.time

#-------------------------------------------------------------------------------
# Aggregate datatypes
#-------------------------------------------------------------------------------
class Array(Schema):
    """
    Matches an array of elements of the same schema
    """
    def __init__(self, raw_element_schema: RawSchema) -> None:
        super().__init__()
        self.element_schema = normalize(raw_element_schema)

    def extract(self, data: Any, path: str, err_ctx: str) -> List:
        if not isinstance(data, list):
            raise SchemaException(f"{err_ctx}: Expected array. Got {type(data).__name__}")

        array = []
        for i, e in enumerate(data):
            element = self.element_schema.extract(e, path, f"{err_ctx}[{i}]")
            array.append(element)
        return array

class FixedMapping(Schema):
    """
    Matches a fixed mapping of specified key:value pairs.

    If a key is not provided, it will be guaranteed to exist and will fall back
    to its default value.
    """
    def __init__(self, schema: Dict[str, RawSchema]) -> None:
        super().__init__()
        self.schema = {}
        for k,v in schema.items():
            self.schema[k] = normalize(v)

    def extract(self, data: Any, path: str, err_ctx: str) -> Dict[str, Any]:
        if not isinstance(data, dict):
            raise SchemaException(f"{err_ctx}: Expected mapping. Got {type(data).__name__}")

        mapping = {} # type: Dict[str, Any]
        for key, schema in self.schema.items():
            if key not in data:
                # Not specified. Assign default
                if isinstance(schema, Array):
                    # Arrays are empty if unspecified
                    mapping[key] = []
                elif isinstance(schema, UserMapping):
                    # User mappings have no entries
                    mapping[key] = {}
                elif isinstance(schema, FixedMapping):
                    # Fixed mappings have all keys populated with their defaults
                    mapping[key] = schema.extract({}, path, f"{err_ctx}.{key}")
                else:
                    mapping[key] = None
            else:
                mapping[key] = schema.extract(data[key], path, f"{err_ctx}.{key}")
        return mapping

class UserMapping(Schema):
    """
    Matches a user-defined mapping. Keys can be anything
    """
    def __init__(self, raw_value_schema: RawSchema) -> None:
        super().__init__()
        self.value_schema = normalize(raw_value_schema)

    def extract(self, data: Any, path: str, err_ctx: str) -> Dict[str, Any]:
        if not isinstance(data, dict):
            raise SchemaException(f"{err_ctx}: Expected mapping. Got {type(data).__name__}")

        mapping = {}
        for key, value in data.items():
            mapping[key] = self.value_schema.extract(value, path, f"{err_ctx}.{key}")
        return mapping


def normalize(raw_schema: RawSchema) -> Schema:
    """
    Convert a raw user-defined scheme into a fully resolved object tree.
    Shorthand representations of aggregate datatypes are expanded into their
    actual classes
    """
    if isinstance(raw_schema, Schema):
        # pass through. no conversion needed
        return raw_schema

    if isinstance(raw_schema, list):
        # Shorthand array. Convert to Array schema object
        if len(raw_schema) != 1:
            raise RuntimeError("Array schema can only have one element")
        return Array(raw_schema[0])

    if isinstance(raw_schema, dict):
        if "*" in raw_schema:
            # Is a user mapping
            if len(raw_schema) != 1:
                raise RuntimeError("User mapping schema shall have only one entry")
            return UserMapping(raw_schema['*'])
        else:
            # Is a fixed mapping
            return FixedMapping(raw_schema)


    raise RuntimeError(f"Invalid schema object: {repr(raw_schema)}")


#-------------------------------------------------------------------------------
# High-level elements
#-------------------------------------------------------------------------------
class Path(String):
    """
    Matches any path
    """
    def __init__(self, shall_exist: bool = True) -> None:
        super().__init__()
        self.shall_exist = shall_exist

    def extract(self, data: Any, path: str, err_ctx: str) -> Any:
        s = super().extract(data, path, err_ctx)
        s = os.path.expanduser(s)

        if not os.path.isabs(s):
            # relative path.
            # extend it to be relative to the config file
            dir_path = os.path.dirname(path)
            s = os.path.join(dir_path, s)

        s = os.path.abspath(s)
        s = os.path.normpath(s)

        if self.shall_exist and not os.path.exists(s):
            raise SchemaException(f"{err_ctx}: Path does not exist: {s}")


        return s

class FilePath(Path):
    """
    Matches a path to a file
    """
    def extract(self, data: Any, path: str, err_ctx: str) -> Any:
        s = super().extract(data, path, err_ctx)
        if self.shall_exist and not os.path.isfile(s):
            raise SchemaException(f"{err_ctx}: Path does not point to a file: {s}")
        return s

class DirectoryPath(Path):
    """
    Matches a path to a directory
    """
    def extract(self, data: Any, path: str, err_ctx: str) -> Any:
        s = super().extract(data, path, err_ctx)
        if self.shall_exist and not os.path.isdir(s):
            raise SchemaException(f"{err_ctx}: Path does not point to a directory: {s}")
        return s

class PythonObjectImport(String):
    """
    Matches a string that specifies a Python object to import.
    For example: ``"my.module.path:ObjectName"``
    """
    def extract(self, data: Any, path: str, err_ctx: str) -> Any:
        s = super().extract(data, path, err_ctx)
        m = re.fullmatch(r"(\w+(?:\.\w+)*):(\w+)", s)
        if not m:
            raise SchemaException(f"{err_ctx}: Invalid object import spec: {s}")

        module_name = m.group(1)
        object_name = m.group(2)

        try:
            module = importlib.import_module(module_name)
        except ModuleNotFoundError as e:
            raise SchemaException(f"{err_ctx}: {str(e)}") from e

        try:
            cls = getattr(module, object_name)
        except AttributeError as e:
            raise SchemaException(f"{err_ctx}: {str(e)}") from e

        return cls



from typing import List, TYPE_CHECKING, Optional
import inspect

from .entry_points import get_entry_points, get_name_from_dist
from ..subcommand import ExporterSubcommand

if TYPE_CHECKING:
    from ..config.loader import AppConfig

class ExporterSubcommandPlugin(ExporterSubcommand):
    """
    Base class for exporter subcommand plugins.
    """
    def __init__(self, dist_name: Optional[str]=None, dist_version: Optional[str]=None) -> None:
        super().__init__()
        self.dist_name = dist_name
        self.dist_version = dist_version

    @property
    def plugin_info(self) -> str:
        if self.dist_name and self.dist_version:
            return f"{self.name} --> {self.dist_name} {self.dist_version}"
        else:
            return f"{self.name} --> {inspect.getabsfile(type(self))}:{type(self).__name__}"


def get_exporter_plugins(cfg: 'AppConfig') -> List[ExporterSubcommandPlugin]:
    """
    Load any plugins that advertise themselves in their setup.py via the following:

    setup(
        ...
        entry_points = {
            "peakrdl.exporters": [
                'my_exporter_name = module.path.to:MyExporter'
            ]
        },
    )
    """
    exporters = []

    # Get exporter plugins from entry-points
    for ep, dist in get_entry_points("peakrdl.exporters"):
        cls = ep.load()
        dist_name = get_name_from_dist(dist)

        if issubclass(cls, ExporterSubcommandPlugin):
            # Override name - always use entry point's name
            cls.name = ep.name
            exporter = cls(dist_name=dist_name, dist_version=dist.version)
        else:
            raise RuntimeError(f"Exporter class {cls} is expected to be extended from peakrdl.plugins.exporter.ExporterSubcommandPlugin")
        exporters.append(exporter)

    # Get any additional exporter plugins from config
    for name, cls in cfg.peakrdl_cfg['plugins']['exporters'].items():
        if issubclass(cls, ExporterSubcommandPlugin):
            # Override name - always use entry point's name
            cls.name = name
            exporter = cls()
        else:
            raise RuntimeError(f"Exporter class {cls} is expected to be extended from peakrdl.plugins.exporter.ExporterSubcommandPlugin")
        exporters.append(exporter)

    return exporters



import os
import setuptools

with open("README.md", "r", encoding='utf-8') as f:
    long_description = f.read()


with open(os.path.join("src/peakrdl", "__about__.py"), "r", encoding='utf-8') as f:
    v_dict = {}
    exec(f.read(), v_dict)
    version = v_dict['__version__']

setuptools.setup(
    name="peakrdl",
    version=version,
    author="Alex Mykyta",
    author_email="amykyta3@github.com",
    description="Command-line tool for control/status register automation",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/SystemRDL/PeakRDL",
    packages=setuptools.find_packages("src"),
    package_dir={"": "src"},
    include_package_data=True,
    entry_points = {
        "console_scripts": ['peakrdl = peakrdl.main:main']
    },
    python_requires='>=3.6',
    install_requires=[
        "systemrdl-compiler >= 1.25.3",
        "peakrdl-html >= 2.10.0",
        "peakrdl-ipxact >= 3.4.0",
        "peakrdl-regblock >= 0.11.0",
        "peakrdl-systemrdl >= 0.2.0",
        "peakrdl-uvm >= 2.3.0",
        "tomli;python_version<'3.11'"
    ],
    classifiers=(
        "Development Status :: 4 - Beta",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3 :: Only",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Electronic Design Automation (EDA)",
    ),
    project_urls={
        "Documentation": "https://peakrdl.readthedocs.io",
        "Source": "https://github.com/SystemRDL/PeakRDL",
        "Tracker": "https://github.com/SystemRDL/PeakRDL/issues",
    },
)



from typing import Optional, List, Any, Dict
import os
import sys

from . import schema

if sys.version_info[0:2] < (3, 11):
    # Prior to py3.11, tomllib is a 3rd party package
    import tomli as tomllib
else:
    # py3.11 and onwards, tomli was absorbed into the standard library as tomllib
    import tomllib

class AppConfig:
    def __init__(self, path: str, raw_data: Dict[str, Any]) -> None:
        self.path = path
        self.raw_data = raw_data

        sch = schema.normalize({
            "plugins": {
                "importers": {"*": schema.PythonObjectImport()},
                "exporters": {"*": schema.PythonObjectImport()},
            },
        })
        self.peakrdl_cfg = self.get_namepsace("peakrdl", sch)

    def get_namepsace(self, name: str, sch: schema.Schema) -> Dict[str, Any]:
        data = self.raw_data.get(name, {})
        try:
            cfg = sch.extract(data, self.path, name)
        except schema.SchemaException as e:
            print(f"{self.path}: error: {str(e)}")
            sys.exit(1)
        return cfg


def _discover_cfg_file() -> Optional[str]:
    """
    Discover a PeakRDL TOML config file
    """

    # 1. In current directory
    path = os.path.abspath("peakrdl.toml")
    if os.path.isfile(path):
        return path

    # 2.
    path = os.path.abspath(".peakrdl.toml")
    if os.path.isfile(path):
        return path

    # 3. Via environment variable
    if "PEAKRDL_CFG" in os.environ:
        path = os.environ["PEAKRDL_CFG"]
        if os.path.isfile(path):
            return path

    # 4. In home directory
    path = os.path.expanduser("~/.peakrdl.toml")
    if os.path.isfile(path):
        return path

    # 5.
    path = os.path.expanduser("~/.config/peakrdl.toml")
    if os.path.isfile(path):
        return path

    # 6. In /etc/
    path = "/etc/peakrdl.toml"
    if os.path.isfile(path):
        return path

    return None


BOOTSTRAP_SCHEMA = schema.normalize({
    "peakrdl": {
        "python_search_paths": [schema.DirectoryPath(shall_exist=False)]
    }
})

def load_cfg(argv: List[str]) -> AppConfig:
    # lazy-parse argv to see if user provided a config file explicitly
    path = None
    argv_iter = iter(argv)
    for arg in argv_iter:
        if arg == "--peakrdl-cfg":
            try:
                path = next(argv_iter)
            except StopIteration:
                print("error: argument --peakrdl-cfg: expected FILE", file=sys.stderr)
                sys.exit(1)
            break

    if path is None:
        # No config file path was provided from the command-line
        # try to discover it elsewhere
        path = _discover_cfg_file()

    if path is None:
        # Nope. Still no config file. Provide empty data
        raw_data = {}
    else:
        if not os.path.isfile(path):
            print(f"error: invalid config file path: {path}", file=sys.stderr)
            sys.exit(1)

        with open(path, 'r', encoding='utf-8') as f:
            s = f.read()
        try:
            raw_data = tomllib.loads(s)
        except tomllib.TOMLDecodeError as e:
            print(f"{path}: error: {str(e)}", file=sys.stderr)
            sys.exit(1)

    # Do a first-pass extraction to fetch additional entries to be added to PYTHONPATH
    try:
        tmp = BOOTSTRAP_SCHEMA.extract(raw_data, path, "")
    except schema.SchemaException as e:
        print(f"{path}: error: {str(e)}")
        sys.exit(1)
    for spath in tmp['peakrdl']['python_search_paths']:
        sys.path.append(spath)

    return AppConfig(path, raw_data)
